/* tcldom-libxml.c --
 *
 *	A Tcl wrapper for libxml's node tree API,
 *	conformant to the TclDOM API.
 *
 * Copyright (c) 2001,2002 Zveno Pty Ltd
 * http://www.zveno.com/
 *
 * Zveno Pty Ltd makes this software and associated documentation
 * available free of charge for any purpose.  You may make copies
 * of the software but you must include all of this notice on any copy.
 *
 * Zveno Pty Ltd does not warrant that this software is error free
 * or fit for any purpose.  Zveno Pty Ltd disclaims any liability for
 * all claims, expenses, losses, damages and costs any user may incur
 * as a result of using, copying or modifying the software.
 *
 * $Id: tcldom-libxml2.c,v 1.17 2002/04/11 21:02:40 balls Exp $
 */

#include "tcldom.h"
#include <libxml/xpath.h>
#include <libxml/xmlIO.h>

#undef TCL_STORAGE_CLASS
#define TCL_STORAGE_CLASS DLLEXPORT

/*
 * Data structures for internal functions.
 */

typedef struct ParserClientData {
  Tcl_Interp *interp;
  Tcl_Obj *externalentityloader;	/* Callback for loading an external entity */
} ParserClientData;

/*
 * Prototypes for procedures defined later in this file:
 */

Tcl_FreeInternalRepProc	TclDOM_DocFree;
Tcl_DupInternalRepProc	TclDOM_DocDup;
Tcl_UpdateStringProc	TclDOM_DocUpdate;
Tcl_SetFromAnyProc	TclDOM_DocSetFromAny;

Tcl_FreeInternalRepProc	TclDOM_NodeFree;
Tcl_DupInternalRepProc	TclDOM_NodeDup;
Tcl_UpdateStringProc	TclDOM_NodeUpdate;
Tcl_SetFromAnyProc	TclDOM_NodeSetFromAny;

Tcl_FreeInternalRepProc	TclDOM_EventFree;
Tcl_DupInternalRepProc	TclDOM_EventDup;
Tcl_UpdateStringProc	TclDOM_EventUpdate;
Tcl_SetFromAnyProc	TclDOM_EventSetFromAny;

/*
 * Forward declarations for private functions.
 */

static void TclDOMGenericError _ANSI_ARGS_((void *ctx, const char *msg, ...));
static Tcl_Obj * TclDOMNewDoc _ANSI_ARGS_((Tcl_Interp *interp));

Tcl_Obj * TclDOM_CreateObjFromNode _ANSI_ARGS_((xmlNodePtr nodePtr));
Tcl_Obj * TclDOM_CreateObjFromDoc  _ANSI_ARGS_((xmlDocPtr  docPtr));


/*
 * Forward declarations of commands
 */

static int TclDOMDOMImplementationCommand _ANSI_ARGS_((ClientData dummy,
						       Tcl_Interp *interp,
						       int objc,
						       Tcl_Obj *CONST objv[]));
static int TclDOMDocumentCommand _ANSI_ARGS_((ClientData dummy,
					      Tcl_Interp *interp,
					      int objc,
					      Tcl_Obj *CONST objv[]));
static int TclDOMNodeCommand _ANSI_ARGS_((ClientData dummy,
					  Tcl_Interp *interp,
					  int objc,
					  Tcl_Obj *CONST objv[]));
static int TclDOMElementCommand _ANSI_ARGS_((ClientData dummy,
					     Tcl_Interp *interp,
					     int objc,
					     Tcl_Obj *CONST objv[]));
static int TclDOMEventCommand _ANSI_ARGS_((ClientData dummy,
					   Tcl_Interp *interp,
					   int objc,
					   Tcl_Obj *CONST objv[]));

/*
 * Functions that implement the TclDOM_Implementation interface
 */

static int TclDOMCreateCommand _ANSI_ARGS_((ClientData dummy,
					    Tcl_Interp *interp,
					    int objc,
					    Tcl_Obj *CONST objv[]));
static int TclDOMParseCommand _ANSI_ARGS_((ClientData dummy,
					    Tcl_Interp *interp,
					    int objc,
					    Tcl_Obj *CONST objv[]));
static int TclDOMSerializeCommand _ANSI_ARGS_((ClientData dummy,
					    Tcl_Interp *interp,
					    int objc,
					    Tcl_Obj *CONST objv[]));
static int TclDOMSelectNodeCommand _ANSI_ARGS_((ClientData dummy,
					    Tcl_Interp *interp,
					    int objc,
					    Tcl_Obj *CONST objv[]));

/*
 * Functions for event support
 */

static Tcl_Obj * TclDOMNewEvent _ANSI_ARGS_((Tcl_Interp *interp,
					     Tcl_Obj *type));

/*
 * Other utilities
 */

static Tcl_Obj * TclDOMGetPath _ANSI_ARGS_((Tcl_Interp *interp,
					    xmlNodePtr nodePtr));

/*
 * libxml2 callbacks
 */

static xmlParserInputPtr TclDOMExternalEntityLoader _ANSI_ARGS_((const char *URL,
								 const char *ID,
								 xmlParserCtxtPtr ctxt));

/*
 * Private libxml2 functions
 */

EXTERN xmlParserCtxtPtr xmlCreateMemoryParserCtxt _ANSI_ARGS_((const char *buffer,
							      int size));

Tcl_ObjType TclDOM_DocObjType = {
  "libxml2-doc",
  TclDOM_DocFree,
  TclDOM_DocDup,
  TclDOM_DocUpdate,
  TclDOM_DocSetFromAny
};

Tcl_ObjType TclDOM_NodeObjType = {
  "libxml2-node",
  TclDOM_NodeFree,
  TclDOM_NodeDup,
  TclDOM_NodeUpdate,
  TclDOM_NodeSetFromAny
};

Tcl_ObjType TclDOM_EventObjType = {
  "tcldom-event",
  TclDOM_EventFree,
  TclDOM_EventDup,
  TclDOM_EventUpdate,
  TclDOM_EventSetFromAny
};

/*
 * Hash tables for mapping string reps to doc/node structure.
 */

static Tcl_HashTable documents;
static Tcl_HashTable nodes;
static int docCntr = 0;
static int nodeCntr = 0;

/*
 * and another for mapping string rep of events to data structure.
 */

static Tcl_HashTable events;
static int eventCntr = 0;

/*
 * Event support.
 *
 * Each table is indexed by xmlNodePtr.
 * The value of an entry is a pointer to a hash table
 * for that node containing listeners, indexed by
 * event type.
 */

static Tcl_HashTable captureListeners;
static Tcl_HashTable bubbleListeners;

/*
 * default loader is overridden while parsing
 */

static xmlExternalEntityLoader defaultLoader;

/*
 * Data structures to support Events
 */

typedef struct TclDOMEvent {
  Tcl_Interp *interp;

  int stopPropagation;
  int preventDefault;
  int dispatched;

  Tcl_Obj *altKey;
  Tcl_Obj *attrName;
  Tcl_Obj *bubbles;
  Tcl_Obj *button;
  Tcl_Obj *cancelable;
  Tcl_Obj *clientX;
  Tcl_Obj *clientY;
  Tcl_Obj *ctrlKey;
  Tcl_Obj *currentNode;
  Tcl_Obj *detail;
  Tcl_Obj *eventPhase;
  Tcl_Obj *metaKey;
  Tcl_Obj *newValue;
  Tcl_Obj *prevValue;
  Tcl_Obj *relatedNode;
  Tcl_Obj *screenX;
  Tcl_Obj *screenY;
  Tcl_Obj *shiftKey;
  Tcl_Obj *target;
  Tcl_Obj *timeStamp;
  Tcl_Obj *type;
  Tcl_Obj *view;
} TclDOMEvent;

/*
 * Accessor function for event objects
 */

static int TclDOM_GetEventFromObj _ANSI_ARGS_((Tcl_Interp *interp,
					       Tcl_Obj *objPtr,
					       TclDOMEvent **eventPtr));

/*
 * For debugging
 */

static Tcl_Channel stderrChan;
static char dbgbuf[200];

/*
 * Switch tables
 */

static char *DOMImplementationCommandMethods[] = {
  "hasFeature",
  "createDocument",
  "create",
  "createDocumentType",
  "createNode",
  "destroy",
  "isNode",
  "parse",
  "selectNode",
  "serialize",
  "trim",
  (char *) NULL
};
enum DOMImplementationCommandMethods {
  TCLDOM_IMPL_HASFEATURE,
  TCLDOM_IMPL_CREATEDOCUMENT,
  TCLDOM_IMPL_CREATE,
  TCLDOM_IMPL_CREATEDOCUMENTYPE,
  TCLDOM_IMPL_CREATENODE,
  TCLDOM_IMPL_DESTROY,
  TCLDOM_IMPL_ISNODE,
  TCLDOM_IMPL_PARSE,
  TCLDOM_IMPL_SELECTNODE,
  TCLDOM_IMPL_SERIALIZE,
  TCLDOM_IMPL_TRIM
};
static char *DocumentCommandMethods[] = {
  "cget",
  "configure",
  "createElement",
  "createDocumentFragment",
  "createTextNode",
  "createComment",
  "createCDATASection",
  "createProcessingInstruction",
  "createAttribute",
  "createEntity",
  "createEntityReference",
  "createDocTypeDecl",
  "importNode",
  "createElementNS",
  "createAttributeNS",
  "getElementsByTagNameNS",
  "getElementsById",
  "createEvent",
  "getElementsByTagName",
  (char *) NULL
};
enum DocumentCommandMethods {
  TCLDOM_DOCUMENT_CGET,
  TCLDOM_DOCUMENT_CONFIGURE,
  TCLDOM_DOCUMENT_CREATEELEMENT,
  TCLDOM_DOCUMENT_CREATEDOCUMENTFRAGMENT,
  TCLDOM_DOCUMENT_CREATETEXTNODE,
  TCLDOM_DOCUMENT_CREATECOMMENT,
  TCLDOM_DOCUMENT_CREATECDATASECTION,
  TCLDOM_DOCUMENT_CREATEPI,
  TCLDOM_DOCUMENT_CREATEATTRIBUTE,
  TCLDOM_DOCUMENT_CREATEENTITY,
  TCLDOM_DOCUMENT_CREATEENTITYREFERENCE,
  TCLDOM_DOCUMENT_CREATEDOCTYPEDECL,
  TCLDOM_DOCUMENT_IMPORTNODE,
  TCLDOM_DOCUMENT_CREATEELEMENTNS,
  TCLDOM_DOCUMENT_CREATEATTRIBUTENS,
  TCLDOM_DOCUMENT_GETELEMENTSBYTAGNAMENS,
  TCLDOM_DOCUMENT_GETELEMENTSBYID,
  TCLDOM_DOCUMENT_CREATEEVENT,
  TCLDOM_DOCUMENT_GETELEMENTSBYTAGNAME
};
static char *DocumentCommandOptions[] = {
  "-doctype",
  "-implementation",
  "-documentElement",
  (char *) NULL
};
enum DocumentCommandOptions {
  TCLDOM_DOCUMENT_DOCTYPE,
  TCLDOM_DOCUMENT_IMPLEMENTATION,
  TCLDOM_DOCUMENT_DOCELEMENT
};
static char *NodeCommandMethods[] = {
  "cget",
  "configure",
  "insertBefore",
  "replaceChild",
  "removeChild",
  "appendChild",
  "hasChildNodes",
  "cloneNode",
  "children",
  "parent",
  "path",
  "createNode",
  "selectNode",
  "stringValue",
  "addEventListener",
  "removeEventListener",
  "dispatchEvent",
  (char *) NULL
};
enum NodeCommandMethods {
  TCLDOM_NODE_CGET,
  TCLDOM_NODE_CONFIGURE,
  TCLDOM_NODE_INSERTBEFORE,
  TCLDOM_NODE_REPLACECHILD,
  TCLDOM_NODE_REMOVECHILD,
  TCLDOM_NODE_APPENDCHILD,
  TCLDOM_NODE_HASCHILDNODES,
  TCLDOM_NODE_CLONENODE,
  TCLDOM_NODE_CHILDREN,
  TCLDOM_NODE_PARENT,
  TCLDOM_NODE_PATH,
  TCLDOM_NODE_CREATENODE,
  TCLDOM_NODE_SELECTNODE,
  TCLDOM_NODE_STRINGVALUE,
  TCLDOM_NODE_ADDEVENTLISTENER,
  TCLDOM_NODE_REMOVEEVENTLISTENER,
  TCLDOM_NODE_DISPATCHEVENT
};
static char *NodeCommandOptions[] = {
  "-nodeType",
  "-parentNode",
  "-childNodes",
  "-firstChild",
  "-lastChild",
  "-previousSibling",
  "-nextSibling",
  "-attributes",
  "-namespaceURI",
  "-prefix",
  "-localName",
  "-nodeValue",
  "-cdatasection",
  "-nodeName",
  "-ownerDocument",
  (char *) NULL
};
enum NodeCommandOptions {
  TCLDOM_NODE_NODETYPE,
  TCLDOM_NODE_PARENTNODE,
  TCLDOM_NODE_CHILDNODES,
  TCLDOM_NODE_FIRSTCHILD,
  TCLDOM_NODE_LASTCHILD,
  TCLDOM_NODE_PREVIOUSSIBLING,
  TCLDOM_NODE_NEXTSIBLING,
  TCLDOM_NODE_ATTRIBUTES,
  TCLDOM_NODE_NAMESPACEURI,
  TCLDOM_NODE_PREFIX,
  TCLDOM_NODE_LOCALNAME,
  TCLDOM_NODE_NODEVALUE,
  TCLDOM_NODE_CDATASECTION,
  TCLDOM_NODE_NODENAME,
  TCLDOM_NODE_OWNERDOCUMENT
};
static char *NodeCommandAddEventListenerOptions[] = {
  "-usecapture",
  (char *) NULL
};
enum NodeCommandAddEventListenerOptions {
  TCLDOM_NODE_ADDEVENTLISTENER_USECAPTURE
};
static char *ElementCommandMethods[] = {
  "cget",
  (char *) NULL
};
enum ElementCommandMethods {
  TCLDOM_ELEMENT_CGET
};
static char *EventCommandMethods[] = {
  "cget",
  "configure",
  "stopPropagation",
  "preventDefault",
  "initEvent",
  "initUIEvent",
  "initMouseEvent",
  "initMutationEvent",
  "postUIEvent",
  "postMouseEvent",
  "postMutationEvent",
  (char *) NULL
};
enum EventCommandMethods {
  TCLDOM_EVENT_CGET,
  TCLDOM_EVENT_CONFIGURE,
  TCLDOM_EVENT_STOPPROPAGATION,
  TCLDOM_EVENT_PREVENTDEFAULT,
  TCLDOM_EVENT_INITEVENT,
  TCLDOM_EVENT_INITUIEVENT,
  TCLDOM_EVENT_INITMOUSEEVENT,
  TCLDOM_EVENT_INITMUTATIONEVENT,
  TCLDOM_EVENT_POSTUIEVENT,
  TCLDOM_EVENT_POSTMOUSEEVENT,
  TCLDOM_EVENT_POSTMUTATIONEVENT
};
static char *EventCommandOptions[] = {
  "-altKey",
  "-attrName",
  "-bubbles",
  "-button",
  "-cancelable",
  "-clientX",
  "-clientY",
  "-ctrlKey",
  "-currentNode",
  "-detail",
  "-eventPhase",
  "-metaKey",
  "-newValue",
  "-prevValue",
  "-relatedNode",
  "-screenX",
  "-screenY",
  "-shiftKey",
  "-target",
  "-timeStamp",
  "-type",
  "-view",
  (char *) NULL
};
enum EventCommandOptions {
  TCLDOM_EVENT_ALTKEY,
  TCLDOM_EVENT_ATTRNAME,
  TCLDOM_EVENT_BUBBLES,
  TCLDOM_EVENT_BUTTON,
  TCLDOM_EVENT_CANCELABLE,
  TCLDOM_EVENT_CLIENTX,
  TCLDOM_EVENT_CLIENTY,
  TCLDOM_EVENT_CTRLKEY,
  TCLDOM_EVENT_CURRENTNODE,
  TCLDOM_EVENT_DETAIL,
  TCLDOM_EVENT_EVENTPHASE,
  TCLDOM_EVENT_METAKEY,
  TCLDOM_EVENT_NEWVALUE,
  TCLDOM_EVENT_PREVVALUE,
  TCLDOM_EVENT_RELATEDNODE,
  TCLDOM_EVENT_SCREENX,
  TCLDOM_EVENT_SCREENY,
  TCLDOM_EVENT_SHIFTKEY,
  TCLDOM_EVENT_TARGET,
  TCLDOM_EVENT_TIMESTAMP,
  TCLDOM_EVENT_TYPE,
  TCLDOM_EVENT_VIEW
};
static char *EventTypes[] = {
  "DOMFocusIn",
  "DOMFocusOut",
  "DOMActivate",
  "click",
  "mousedown",
  "mouseup",
  "mouseover",
  "mousemove",
  "mouseout"
  "DOMSubtreeModified",
  "DOMNodeInserted",
  "DOMNodeRemoved",
  "DOMNodeInsertedIntoDocument",
  "DOMNodeRemovedFromDocument",
  "DOMAttrModified",
  "DOMAttrRemoved",
  "DOMCharacterDataModified"
};
enum EventTypes {
  TCLDOM_EVENT_DOMFOCUSIN,
  TCLDOM_EVENT_DOMFOCUSOUT,
  TCLDOM_EVENT_DOMACTIVATE,
  TCLDOM_EVENT_CLICK,
  TCLDOM_EVENT_MOUSEDOWN,
  TCLDOM_EVENT_MOUSEUP,
  TCLDOM_EVENT_MOUSEOVER,
  TCLDOM_EVENT_MOUSEMOVE,
  TCLDOM_EVENT_MOUSEOUT,
  TCLDOM_EVENT_DOMSUBTREEMODIFIED,
  TCLDOM_EVENT_DOMNODEINSERTED,
  TCLDOM_EVENT_DOMNODEREMOVED,
  TCLDOM_EVENT_DOMNODEINSERTEDINTODOCUMENT,
  TCLDOM_EVENT_DOMNODEREMOVEDFROMDOCUMENT,
  TCLDOM_EVENT_DOMATTRMODIFIED,
  TCLDOM_EVENT_DOMATTRREMOVED,
  TCLDOM_EVENT_DOMCHARACTERDATAMODIFIED
};
static char *ParseCommandOptions[] = {
  "-baseuri",
  "-externalentitycommand",
  (char *) NULL
};
enum ParseCommandOptions {
  TCLDOM_PARSE_BASEURI,
  TCLDOM_PARSE_EXTERNALENTITYCOMMAND
};

/*
 * Error context for passing error result back to caller.
 */

typedef struct GenericError_Info {
  Tcl_Interp *interp;
  int code;
  Tcl_Obj *msg;
} GenericError_Info;

static GenericError_Info errorInfo = {
  NULL,
  TCL_OK,
  NULL
};

/*
 * Default values
 */

EXTERN int xmlLoadExtDtdDefaultValue;


/*
 *----------------------------------------------------------------------------
 *
 * Tcldomxml_Init --
 *
 *  Initialisation routine for loadable module
 *
 * Results:
 *  None.
 *
 * Side effects:
 *  Creates commands in the interpreter,
 *
 *----------------------------------------------------------------------------
 */

int
Tcldomxml_Init (interp)
     Tcl_Interp *interp;	/* Interpreter to initialise */
{
  TclDOM_Implementation *implPtr;
  int dbgMode;

  Tcl_InitStubs(interp, "8.1", 0);

  /* Configure the libxml2 parser */
  xmlInitMemory();
  xmlSubstituteEntitiesDefault(1);

  /*
   * 6 will load external entities.
   * 0 will not.
   * TODO: provide configuration option for setting this value.
   */
  xmlLoadExtDtdDefaultValue = 6;

  xmlSetGenericErrorFunc((void *) &errorInfo, TclDOMGenericError);

  xmlXPathInit();

  Tcl_RegisterObjType(&TclDOM_DocObjType);
  Tcl_RegisterObjType(&TclDOM_NodeObjType);

  /*
   * Register this DOM implementation with the TclDOM
   * generic layer.  We must do this for each separate
   * Tcl object type.
   */

  implPtr = (TclDOM_Implementation *) Tcl_Alloc(sizeof(TclDOM_Implementation));
  implPtr->name = Tcl_NewStringObj("libxml2-doc", -1);
  Tcl_IncrRefCount(implPtr->name);
  implPtr->type = &TclDOM_DocObjType;
  implPtr->create = TclDOMCreateCommand;
  implPtr->parse = TclDOMParseCommand;
  implPtr->serialize = TclDOMSerializeCommand;
  implPtr->document = TclDOMDocumentCommand;
  implPtr->documentfragment = NULL;
  implPtr->node = TclDOMNodeCommand;
  implPtr->element = NULL;
  TclDOM_Register(interp, implPtr);

  implPtr = (TclDOM_Implementation *) Tcl_Alloc(sizeof(TclDOM_Implementation));
  implPtr->name = Tcl_NewStringObj("libxml2-node", -1);
  Tcl_IncrRefCount(implPtr->name);
  implPtr->type = &TclDOM_NodeObjType;
  implPtr->create = TclDOMCreateCommand;
  implPtr->parse = TclDOMParseCommand;
  implPtr->serialize = TclDOMSerializeCommand;
  implPtr->document = TclDOMDocumentCommand;
  implPtr->documentfragment = NULL;
  implPtr->node = TclDOMNodeCommand;
  implPtr->element = TclDOMElementCommand;
  TclDOM_Register(interp, implPtr);

  Tcl_CreateObjCommand(interp, "dom::libxml2::DOMImplementation", TclDOMDOMImplementationCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::document", TclDOMDocumentCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::node", TclDOMNodeCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::create", TclDOMCreateCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::parse", TclDOMParseCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::serialize", TclDOMSerializeCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::selectnode", TclDOMSelectNodeCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::selectNode", TclDOMSelectNodeCommand, NULL, NULL);
/*
  Tcl_CreateObjCommand(interp, "dom::libxml2::element", TclDOMElementCommand, NULL, NULL);
*/
  Tcl_CreateObjCommand(interp, "dom::libxml2::event", TclDOMEventCommand, NULL, NULL);

  Tcl_InitHashTable(&documents, TCL_STRING_KEYS);
  Tcl_InitHashTable(&nodes, TCL_STRING_KEYS);
  Tcl_InitHashTable(&events, TCL_STRING_KEYS);
  Tcl_InitHashTable(&captureListeners, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&bubbleListeners, TCL_ONE_WORD_KEYS);

  stderrChan = Tcl_GetChannel(interp, "stderr", &dbgMode);

  Tcl_PkgProvide(interp, "dom::libxml2", "1.0");

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMCreateCommand --
 *
 *  Implements dom::libxml2::create command
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMCreateCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }

  Tcl_SetObjResult(interp, TclDOMNewDoc(interp));
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMParseCommand --
 *
 *  Implements dom::libxml2::parse command
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMParseCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  char *buf;
  int len, new, option;
  xmlParserCtxtPtr ctxt;
  xmlDocPtr docPtr;
  Tcl_Obj *objPtr;
  Tcl_HashEntry *entryPtr;
  Tcl_Obj *baseuriPtr = NULL;
  Tcl_Obj *extentity = NULL;
  ParserClientData *clientData;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "xml ?args ...?");
    return TCL_ERROR;
  }

  buf = Tcl_GetStringFromObj(objv[1], &len);
  if (buf == NULL) {
    Tcl_SetResult(interp, "unable to get document to parse", NULL);
    return TCL_ERROR;
  }

  /*
   * Process options
   */

  objc -= 2;
  objv += 2;
  while (objc) {
    if (objc == 1) {
      Tcl_Obj *msgPtr;

      msgPtr = Tcl_NewStringObj("missing value for configuration option \"", -1);
      Tcl_AppendObjToObj(msgPtr, objv[0]);
      Tcl_AppendStringsToObj(msgPtr, "\"", (char *) NULL);
      Tcl_SetObjResult(interp, msgPtr);
      return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[0], ParseCommandOptions,
			    "option", 0, &option) != TCL_OK) {
      return TCL_ERROR;
    }

    switch ((enum ParseCommandOptions) option) {
    case TCLDOM_PARSE_BASEURI:
      baseuriPtr = objv[1];
      break;

    case TCLDOM_PARSE_EXTERNALENTITYCOMMAND:
      extentity = objv[1];
      break;
    }

    objc -= 2;
    objv += 2;
  }

  /*
   * This does the real work
   */

  xmlInitParser();
  ctxt = xmlCreateMemoryParserCtxt(buf, len);
  if (ctxt == NULL) {
    /* Out of memory - we're in big trouble... */
    Tcl_SetResult(interp, "unable to allocate parser context", NULL);
    return TCL_ERROR;
  }

  /*
   * Use the _private field to store TclDOM data
   */

  clientData = (ParserClientData *) Tcl_Alloc(sizeof(ParserClientData));
  ctxt->_private = (char *) clientData;
  clientData->interp = interp;

  if (baseuriPtr) {
    ctxt->input->filename = Tcl_GetStringFromObj(baseuriPtr, NULL);
  }

  defaultLoader = xmlGetExternalEntityLoader();
  if (extentity) {
    clientData->externalentityloader = extentity;
    xmlSetExternalEntityLoader(TclDOMExternalEntityLoader);
  }

  /* docPtr = xmlParseMemory(buf, len); */
  if (xmlParseDocument(ctxt) != 0) {
    docPtr = NULL;
  } else {
    docPtr = ctxt->myDoc;
  }

  xmlSetExternalEntityLoader(defaultLoader);
  xmlFreeParserCtxt(ctxt);
  xmlCleanupParser();
  Tcl_Free(ctxt->_private);

  if (docPtr == NULL) {
    Tcl_SetResult(interp, "unable to parse document", NULL);
    return TCL_ERROR;
  }

  /*
   * Make sure base URI is stored in the document.
   * Setting the input filename is insufficient.
   */

  if (baseuriPtr && docPtr->URL == NULL) {
    buf = Tcl_GetStringFromObj(baseuriPtr, &len);
    docPtr->URL = Tcl_Alloc(len + 1);
    strcpy((char *) docPtr->URL, buf);
  }

  /*
   * Wrap the document pointer in a Tcl object
   */

  buf = Tcl_Alloc(20);
  sprintf(buf, "doc%d", docCntr++);

  entryPtr = Tcl_CreateHashEntry(&documents, buf, &new);
  if (!new) {
    Tcl_Free(buf);
    Tcl_SetResult(interp, "internal error: previously allocated token", NULL);
    return TCL_ERROR;
  }
  Tcl_SetHashValue(entryPtr, (void *) docPtr);

  objPtr = Tcl_NewObj();
  objPtr->internalRep.otherValuePtr = (VOID *) docPtr;
  objPtr->typePtr = &TclDOM_DocObjType;
  objPtr->bytes = Tcl_Alloc(20);
  strcpy(objPtr->bytes, buf);
  objPtr->length = strlen(buf);

  Tcl_Free(buf);

  /*
   * Use the _private field to point back to the wrapping
   * object.  This allows convenient reuse of the
   * object.
   */

  docPtr->_private = (void *) objPtr;

  Tcl_SetObjResult(interp, objPtr);

  return TCL_OK;
}

/*
 * Interpose on resolving external entity references.
 *
 * The return code of the script evaluation determines
 * the behaviour:
 * TCL_OK - return value is to be used as the entity data
 * TCL_CONTINUE - use the default libxml2 entity loader
 * TCL_BREAK - forget about this entity
 * TCL_ERROR - background error
 */

static xmlParserInputPtr
TclDOMExternalEntityLoader(url, id, ctxt)
     const char *url;
     const char *id;
     xmlParserCtxtPtr ctxt;
{
  ParserClientData *clientData = (ParserClientData *) ctxt->_private;
  Tcl_Obj *cmdPtr;

  if (clientData) {
    cmdPtr = Tcl_DuplicateObj(clientData->externalentityloader);
  } else {
    return NULL;
  }

  if (url) {
    Tcl_ListObjAppendElement(clientData->interp, cmdPtr, Tcl_NewStringObj(url, -1));
  } else {
    Tcl_ListObjAppendElement(clientData->interp, cmdPtr, Tcl_NewListObj(0, NULL));
  }
  if (id) {
    Tcl_ListObjAppendElement(clientData->interp, cmdPtr, Tcl_NewStringObj(id, -1));
  } else {
    Tcl_ListObjAppendElement(clientData->interp, cmdPtr, Tcl_NewListObj(0, NULL));
  }

  switch (Tcl_GlobalEvalObj(clientData->interp, cmdPtr)) {

  case TCL_OK:
    /* TODO: Use the return value as the input */
    return (defaultLoader)(url, id, ctxt);

  case TCL_CONTINUE:
    /* Use the default libxml2 loader */
    return (defaultLoader)(url, id, ctxt);

  case TCL_BREAK:
    /* Do not load external entity, but no error */
    return NULL;

  default:
    Tcl_BackgroundError(clientData->interp);
    return NULL;
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMSerializeCommand --
 *
 *  Implements dom::libxml2::serialize command
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMSerializeCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  xmlDocPtr docPtr;
  xmlNodePtr nodePtr;
  xmlChar *result;
  int len;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "node ?args ...?");
    return TCL_ERROR;
  }

  if (TclDOM_GetDocFromObj(interp, objv[1], &docPtr) != TCL_OK) {
    if (TclDOM_GetNodeFromObj(interp, objv[1], &nodePtr) == TCL_OK) {
      /* Serialize just the node */
      Tcl_SetResult(interp, "not yet implemented - serialize whole document", NULL);
      return TCL_ERROR;
    } else {
      Tcl_SetResult(interp, "not a libxml2 node", NULL);
      return TCL_ERROR;
    }
  }

  xmlDocDumpMemory(docPtr, &result, &len);
  Tcl_SetObjResult(interp, Tcl_NewStringObj(result, len));
  free(result);

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMDOMImplementationCommand --
 *
 *  Implements dom::libxml2::DOMImplementation command
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMDOMImplementationCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int method;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "method ?args...?");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], DOMImplementationCommandMethods,
			  "method", 0, &method) != TCL_OK) {
    return TCL_ERROR;
  }

  switch ((enum DOMImplementationCommandMethods) method) {
  case TCLDOM_IMPL_CREATE:
    if (objc == 2) {
      return TclDOMCreateCommand(dummy, interp, 0, NULL);
    } else if (objc == 3) {
      Tcl_Obj *objPtr;
      xmlDocPtr docPtr;
      xmlNodePtr nodePtr;

      if (TclDOMCreateCommand(dummy, interp, 0, NULL) != TCL_OK) {
	return TCL_ERROR;
      }
      objPtr = Tcl_GetObjResult(interp);
      TclDOM_GetDocFromObj(interp, objPtr, &docPtr);
      nodePtr = xmlNewDocNode(docPtr, NULL, Tcl_GetStringFromObj(objv[2], NULL), NULL);
      if (nodePtr == NULL) {
	Tcl_SetResult(interp, "unable to create document element", NULL);
	Tcl_DecrRefCount(objPtr);
	return TCL_ERROR;
      }

      Tcl_SetObjResult(interp, objPtr);
    } else {
      Tcl_WrongNumArgs(interp, 1, objv, "create ?doc?");
      return TCL_ERROR;
    }

    break;

  case TCLDOM_IMPL_PARSE:
    return TclDOMParseCommand(dummy, interp, objc, objv);

  case TCLDOM_IMPL_SERIALIZE:
    return TclDOMSerializeCommand(dummy, interp, objc, objv);

  case TCLDOM_IMPL_SELECTNODE:
    return TclDOMSelectNodeCommand(dummy, interp, objc, objv);

  default:
    Tcl_SetResult(interp, "not yet implemented", NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMSelectNodeCommand --
 *
 *  Implements dom::libxml2::selectnode command.
 *
 * Results:
 *  Returns result of XPath expression evaluation.
 *
 * Side effects:
 *  Memory is allocated for Tcl object to return result.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMSelectNodeCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int i, len;
  char *path;
  Tcl_Obj *objPtr;
  xmlDocPtr docPtr;
  xmlNodePtr nodePtr = NULL;
  xmlXPathContextPtr ctxt;
  xmlXPathObjectPtr xpathObj;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "doc location-path");
    return TCL_ERROR;
  }

  path = Tcl_GetStringFromObj(objv[2], &len);
  if (len == 0) {
    return TCL_OK;
  }

  if (TclDOM_GetDocFromObj(interp, objv[1], &docPtr) != TCL_OK) {
    if (TclDOM_GetNodeFromObj(interp, objv[1], &nodePtr) == TCL_OK) {
      docPtr = nodePtr->doc;
    } else {
      return TCL_ERROR;
    }
  }

  ctxt = xmlXPathNewContext(docPtr);
  if (ctxt == NULL) {
    Tcl_SetResult(interp, "unable to create XPath context", NULL);
    return TCL_ERROR;
  }

  if (nodePtr) {
    ctxt->node = nodePtr;
  }

  xpathObj = xmlXPathEval(path, ctxt);
  if (xpathObj == NULL) {
    Tcl_SetResult(interp, "error evaluating XPath location path", NULL);
    goto error;
  }

  objPtr = Tcl_NewObj();
  switch (xpathObj->type) {

  case XPATH_NODESET:
    len = xmlXPathNodeSetGetLength(xpathObj->nodesetval);
    for (i = 0; i < len; i++) {
      nodePtr = xmlXPathNodeSetItem(xpathObj->nodesetval, i);
      Tcl_ListObjAppendElement(interp, objPtr, TclDOM_CreateObjFromNode(nodePtr));
    }
    break;

  case XPATH_BOOLEAN:
    Tcl_SetBooleanObj(objPtr, xpathObj->boolval);
    break;

  case XPATH_NUMBER:
    Tcl_SetDoubleObj(objPtr, xpathObj->floatval);
    break;

  case XPATH_STRING:
    Tcl_SetStringObj(objPtr, xpathObj->stringval, strlen(xpathObj->stringval));
    break;

  default:
    Tcl_SetResult(interp, "bad XPath object type", NULL);
    goto error2;
  }

  xmlXPathFreeObject(xpathObj);
  xmlXPathFreeContext(ctxt);

  Tcl_SetObjResult(interp, objPtr);
  return TCL_OK;

 error2:
  xmlXPathFreeObject(xpathObj);

 error:
  xmlXPathFreeContext(ctxt);
  return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMDocumentCommand --
 *
 *  Implements dom::libxml2::document command.
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMDocumentCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int method, option, postMutationEvent = 0;
  xmlDocPtr docPtr = NULL;
  xmlNodePtr nodePtr = NULL;

  if (objc < 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "method token ?args...?");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], DocumentCommandMethods,
			  "method", 0, &method) != TCL_OK) {
    return TCL_ERROR;
  }

  if (TclDOM_GetDocFromObj(interp, objv[2], &docPtr) != TCL_OK) {
    docPtr = NULL;
    if (TclDOM_GetNodeFromObj(interp, objv[2], &nodePtr) != TCL_OK) {
      return TCL_ERROR;
    }
  }

  switch ((enum DocumentCommandMethods) method) {

  case TCLDOM_DOCUMENT_CGET:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "cget option");
      return TCL_ERROR;
    }

    if (!docPtr) {
      Tcl_SetResult(interp, "not a document", NULL);
    }

    if (Tcl_GetIndexFromObj(interp, objv[3], DocumentCommandOptions,
			    "option", 0, &option) != TCL_OK) {
      return TCL_ERROR;
    }

    switch ((enum DocumentCommandOptions) option) {

    case TCLDOM_DOCUMENT_DOCTYPE:
    case TCLDOM_DOCUMENT_IMPLEMENTATION:
      Tcl_SetResult(interp, "not yet implemented", NULL);
      return TCL_ERROR;

    case TCLDOM_DOCUMENT_DOCELEMENT:

      nodePtr = xmlDocGetRootElement(docPtr);

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(nodePtr));

      break;

    default:
      Tcl_SetResult(interp, "unknown option", NULL);
      return TCL_ERROR;
    }

    break;

  case TCLDOM_DOCUMENT_CONFIGURE:

    if (objc == 4) {
      Tcl_Obj *newobjv[4];

      newobjv[0] = objv[0];
      newobjv[1] = Tcl_NewStringObj("cget", -1);
      newobjv[2] = objv[2];
      newobjv[3] = objv[3];

      if (TclDOMDocumentCommand(NULL, interp, 4, newobjv) != TCL_OK) {
	return TCL_ERROR;
      }

    } else {
      Tcl_SetResult(interp, "options are read-only", NULL);
    }

    break;

  case TCLDOM_DOCUMENT_CREATEELEMENT:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "token name");
      return TCL_ERROR;
    }

    if (docPtr) {
      xmlNodePtr new = xmlNewDocNode(docPtr, NULL, Tcl_GetStringFromObj(objv[3], NULL), NULL);
      xmlNodePtr old = xmlDocSetRootElement(docPtr, new);
      if (old) {
	xmlDocSetRootElement(docPtr, old);
	xmlFreeNode(new);
	Tcl_SetResult(interp, "document element already exists", NULL);
	return TCL_ERROR;
      }
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(new));
    } else {
      xmlNodePtr new = xmlNewChild(nodePtr, NULL, Tcl_GetStringFromObj(objv[3], NULL), NULL);
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(new));
    }

    postMutationEvent = 1;

    break;

  case TCLDOM_DOCUMENT_CREATEDOCUMENTFRAGMENT:

    if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "token");
      return TCL_ERROR;
    }

    if (docPtr) {
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewDocFragment(docPtr)));
    } else {
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewDocFragment(nodePtr->doc)));
    }

    postMutationEvent = 1;

    break;

  case TCLDOM_DOCUMENT_CREATETEXTNODE:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "token text");
      return TCL_ERROR;
    }

    if (docPtr) {
      char *content;
      int len;

      content = Tcl_GetStringFromObj(objv[3], &len);
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewDocTextLen(docPtr, content, len)));
    } else {
      xmlNodePtr textnode, returnNode;
      char *content;
      int len;

      content = Tcl_GetStringFromObj(objv[3], &len);
      textnode = xmlNewTextLen(content, len);
      if (textnode == NULL) {
	Tcl_SetResult(interp, "creating text node failed", NULL);
	return TCL_ERROR;
      }
      returnNode = xmlAddChild(nodePtr, textnode);
      if (returnNode == NULL) {
	xmlFreeNode(textnode);
	Tcl_SetResult(interp, "add child failed", NULL);
	return TCL_ERROR;
      }

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(textnode));
    }

    postMutationEvent = 1;

    break;

  case TCLDOM_DOCUMENT_CREATECOMMENT:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "token data");
      return TCL_ERROR;
    }

    if (docPtr) {
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewComment(Tcl_GetStringFromObj(objv[3], NULL))));
    } else {
      xmlNodePtr commentnode;
      int len;
      xmlChar *content;

      content = Tcl_GetStringFromObj(objv[3], &len);
      commentnode = xmlNewTextLen(content, len);
      xmlAddChild(nodePtr, commentnode);

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(commentnode));
    }

    postMutationEvent = 1;

    break;

  case TCLDOM_DOCUMENT_CREATECDATASECTION:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "token text");
      return TCL_ERROR;
    }

    if (docPtr) {
      char *content;
      int len;

      content = Tcl_GetStringFromObj(objv[3], &len);
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewDocTextLen(docPtr, content, len)));
    } else {
      xmlNodePtr textnode;
      char *content;
      int len;

      content = Tcl_GetStringFromObj(objv[3], &len);
      textnode = xmlNewTextLen(content, len);
      xmlAddChild(nodePtr, textnode);

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(textnode));
    }

    postMutationEvent = 1;

    break;

  case TCLDOM_DOCUMENT_CREATEEVENT:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "token type");
    }

    Tcl_SetObjResult(interp, TclDOMNewEvent(interp, objv[3]));

    break;

  default:
    Tcl_SetResult(interp, "not yet implemented", NULL);
    return TCL_ERROR;
  }

  if (postMutationEvent) {
    Tcl_Obj *newobjv[10];
    int newobjc;

    newobjv[0] = Tcl_NewStringObj("event", -1);
    newobjv[1] = Tcl_NewStringObj("postMutationEvent", -1);
    newobjv[2] = Tcl_GetObjResult(interp);

    newobjv[3] = Tcl_NewStringObj("DOMNodeInserted", -1);
    newobjv[4] = Tcl_NewStringObj("-relatedNode", -1);
    newobjv[5] = objv[2]; /* token */
    newobjv[6] = NULL;

    if (TclDOMEventCommand(NULL, interp, 6, newobjv) != TCL_OK) {
      return TCL_ERROR;
    }

    return TCL_OK;

    newobjv[3] = Tcl_NewStringObj("DOMNodeInsertedIntoDocument", -1);
    newobjv[4] = NULL;

    if (TclDOMEventCommand(NULL, interp, 4, newobjv) != TCL_OK) {
      return TCL_ERROR;
    }

    /*
     * We just added the document element, so don't bother
     * posting an event to the document.
     */

    return TCL_OK;

    if (nodePtr) {
      newobjv[2] = objv[2];
      newobjv[3] = Tcl_NewStringObj("DOMSubtreeModified ", -1);
      newobjv[4] = NULL;

      if (TclDOMEventCommand(NULL, interp, 4, newobjv) != TCL_OK) {
	return TCL_ERROR;
      }
    }
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TriggerEventListeners --
 *
 *  Iterates through the list of event listeners for
 *  a node and fires events.
 *
 * Results:
 *  Depends on listeners.
 *
 * Side effects:
 *  Depends on listeners.
 *
 *----------------------------------------------------------------------------
 */

static int
TriggerEventListeners(interp, type, nodePtr, eventPtr)
     Tcl_Interp *interp;
     Tcl_HashTable *type;
     xmlNodePtr nodePtr;
     Tcl_Obj *eventPtr;
{
  TclDOMEvent *event = (TclDOMEvent *) eventPtr->internalRep.otherValuePtr;
  Tcl_HashEntry *entryPtr;
  Tcl_HashTable *tablePtr;
  Tcl_Obj *listenerListPtr;
  int listenerLen, listenerIdx;

  entryPtr = Tcl_FindHashEntry(type, (char *) nodePtr);
  if (!entryPtr) {
    return TCL_OK;
  }
  tablePtr = (Tcl_HashTable *) Tcl_GetHashValue(entryPtr);

  entryPtr = Tcl_FindHashEntry(tablePtr, Tcl_GetStringFromObj(event->type, NULL));
  if (!entryPtr) {
    return TCL_OK;
  }
  listenerListPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);

  /*
   * DOM L2 specifies that the ancestors are determined
   * at the moment of event dispatch, so using a static
   * list is the correct thing to do.
   */

  Tcl_ListObjLength(interp, listenerListPtr, &listenerLen);
  for (listenerIdx = 0; listenerIdx < listenerLen; listenerIdx++) {
    Tcl_Obj *listenerObj, *cmdPtr;
    Tcl_Obj **objv;
    int objc;

    Tcl_ListObjIndex(interp, listenerListPtr, listenerIdx, &listenerObj);

    /*
     * BUG workaround: eval'ing the command loses the event
     * object's internal rep.  By the time it gets to EventSetFromAny
     * the Tcl object seems to be corrupt.  As a workaround,
     * assume that the listener script is a single command
     * and invoke directly.
     *

    cmdPtr = Tcl_DuplicateObj(listenerObj);
    if (Tcl_ListObjAppendElement(interp, cmdPtr, eventPtr) != TCL_OK) {
      return TCL_ERROR;
    }
    if (Tcl_GlobalEvalObj(interp, cmdPtr) != TCL_OK) {
      Tcl_BackgroundError(interp);
    }
    */

    if (Tcl_ListObjGetElements(interp, listenerObj, &objc, &objv) != TCL_OK) {
      Tcl_BackgroundError(interp);
    } else {
      Tcl_Obj **newobjv = (Tcl_Obj **) Tcl_Alloc((objc + 2) * sizeof(Tcl_Obj *));
      int count;

      for (count = 0; count < objc; count++) {
	newobjv[count] = objv[count];
      }
      newobjv[count] = eventPtr;
      newobjv[count + 1] = NULL;

      if (Tcl_EvalObjv(interp, count + 1, newobjv, TCL_EVAL_GLOBAL) != TCL_OK) {
	Tcl_BackgroundError(interp);
      }
    }

  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMNodeCommand --
 *
 *  Implements dom::libxml2::node command.
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMNodeCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int method, option, len, newobjc, new, usecapture = 0;
  char *buf;
  xmlNodePtr nodePtr = NULL, childNodePtr;
  xmlDocPtr docPtr = NULL;
  Tcl_Obj *nodeObjPtr = NULL;
  Tcl_Obj *docObjPtr = NULL;
  Tcl_Obj *objPtr, *resultPtr, *evPtr;
  Tcl_Obj *newobjv[20];
  Tcl_HashEntry *entryPtr;
  Tcl_HashTable *tablePtr;
  Tcl_Obj *typePtr = objv[3];
  Tcl_Obj *listenerPtr = objv[4];

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "method ?args...?");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], NodeCommandMethods,
			  "method", 0, &method) != TCL_OK) {
    return TCL_ERROR;
  }

  if (TclDOM_GetNodeFromObj(interp, objv[2], &nodePtr) != TCL_OK) {
    if (TclDOM_GetDocFromObj(interp, objv[2], &docPtr) != TCL_OK) {
      return TCL_ERROR;
    } else {
      docObjPtr = objv[2];
      nodePtr = NULL;
    }
  } else {
    nodeObjPtr = objv[2];
    docPtr = NULL;
  }

  switch ((enum NodeCommandMethods) method) {

  case TCLDOM_NODE_CGET:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "cget option");
      return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[3], NodeCommandOptions,
			    "option", 0, &option) != TCL_OK) {
      return TCL_ERROR;
    }

    switch ((enum NodeCommandOptions) option) {

    case TCLDOM_NODE_NODETYPE:

      if (docPtr) {
	Tcl_SetResult(interp, "document", NULL);
	break;
      }

      switch (nodePtr->type) {
      case XML_ELEMENT_NODE:
	Tcl_SetResult(interp, "element", NULL);
	break;
      case XML_ATTRIBUTE_NODE:
	Tcl_SetResult(interp, "attribute", NULL);
	break;
      case XML_TEXT_NODE:
      case XML_CDATA_SECTION_NODE:
	Tcl_SetResult(interp, "textNode", NULL);
	break;
      case XML_ENTITY_REF_NODE:
	Tcl_SetResult(interp, "entityReference", NULL);
	break;
      case XML_ENTITY_NODE:
	Tcl_SetResult(interp, "entity", NULL);
	break;
      case XML_PI_NODE:
	Tcl_SetResult(interp, "processingInstruction", NULL);
	break;
      case XML_COMMENT_NODE:
	Tcl_SetResult(interp, "comment", NULL);
	break;
      case XML_DOCUMENT_NODE:
	Tcl_SetResult(interp, "document", NULL);
	break;
      case XML_DOCUMENT_TYPE_NODE:
	Tcl_SetResult(interp, "docType", NULL);
	break;
      case XML_DOCUMENT_FRAG_NODE:
	Tcl_SetResult(interp, "documentFragment", NULL);
	break;
      case XML_NOTATION_NODE:
	Tcl_SetResult(interp, "notation", NULL);
	break;
      case XML_HTML_DOCUMENT_NODE:
	Tcl_SetResult(interp, "HTMLdocument", NULL);
	break;
      case XML_DTD_NODE:
	Tcl_SetResult(interp, "dtd", NULL);
	break;
      case XML_ELEMENT_DECL:
	Tcl_SetResult(interp, "elementDecl", NULL);
	break;
      case XML_ATTRIBUTE_DECL:
	Tcl_SetResult(interp, "attributeDecl", NULL);
	break;
      case XML_ENTITY_DECL:
	Tcl_SetResult(interp, "entityDecl", NULL);
	break;
      case XML_NAMESPACE_DECL:
	Tcl_SetResult(interp, "namespaceDecl", NULL);
	break;
      case XML_XINCLUDE_START:
	Tcl_SetResult(interp, "xincludeStart", NULL);
	break;
      case XML_XINCLUDE_END:
	Tcl_SetResult(interp, "xincludeEnd", NULL);
	break;
      default:
	Tcl_SetResult(interp, "unknown", NULL);
      }

      break;

    case TCLDOM_NODE_NODENAME:

      if (docPtr) {
	Tcl_SetResult(interp, "#document", NULL);
	break;
      }
      objPtr = Tcl_NewStringObj(nodePtr->name, -1);
      Tcl_SetObjResult(interp, objPtr);

      break;

    case TCLDOM_NODE_NODEVALUE:

      if (docPtr) {
	break;
      }

      if (XML_GET_CONTENT(nodePtr) != NULL) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj(XML_GET_CONTENT(nodePtr), -1));
      }

      break;

    case TCLDOM_NODE_OWNERDOCUMENT:

      if (docPtr) {
	Tcl_SetObjResult(interp, objv[2]);
	break;
      }

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromDoc(nodePtr->doc));

      break;

    case TCLDOM_NODE_PARENTNODE:

      if (docPtr) {
	break;
      }

      if (nodePtr->parent) {
	Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(nodePtr->parent));
      } else {
	Tcl_SetObjResult(interp, TclDOM_CreateObjFromDoc(nodePtr->doc));
      }

      break;

    case TCLDOM_NODE_FIRSTCHILD:

      if (docPtr) {
	childNodePtr = docPtr->children;
      } else {
	childNodePtr = nodePtr->children;
      }

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(childNodePtr));

      break;

    default:
      Tcl_SetResult(interp, "unknown option or not yet implemented", NULL);
      return TCL_ERROR;
    }

    break;

  case TCLDOM_NODE_PATH:

    if (docPtr) {
      newobjv[0] = TclDOM_CreateObjFromDoc(docPtr);
      newobjv[1] = NULL;
      Tcl_SetObjResult(interp, Tcl_NewListObj(1, newobjv));
    } else {
      Tcl_SetObjResult(interp, TclDOMGetPath(interp, nodePtr));
    }

    break;

  case TCLDOM_NODE_CONFIGURE:

    if (objc < 5) {
      Tcl_WrongNumArgs(interp, 3, objv, "configure token option ?value? ?option value ...?");
      return TCL_ERROR;
    }

    if (objc == 4) {
      /* equivalent to cget */
      return TCL_OK;
    }

    objc -= 3;
    objv += 3;
    while (objc) {
      if (objc == 1) {
	Tcl_SetResult(interp, "missing value", NULL);
	return TCL_ERROR;
      }

      if (Tcl_GetIndexFromObj(interp, objv[0], NodeCommandOptions,
			    "option", 0, &option) != TCL_OK) {
	return TCL_ERROR;
      }

      switch ((enum NodeCommandOptions) option) {
      case TCLDOM_NODE_NODETYPE:
      case TCLDOM_NODE_NODENAME:
      case TCLDOM_NODE_PARENTNODE:
      case TCLDOM_NODE_CHILDNODES:
      case TCLDOM_NODE_FIRSTCHILD:
      case TCLDOM_NODE_LASTCHILD:
      case TCLDOM_NODE_PREVIOUSSIBLING:
      case TCLDOM_NODE_NEXTSIBLING:
      case TCLDOM_NODE_ATTRIBUTES:
      case TCLDOM_NODE_NAMESPACEURI:
      case TCLDOM_NODE_PREFIX:
      case TCLDOM_NODE_LOCALNAME:
      case TCLDOM_NODE_OWNERDOCUMENT:

	Tcl_SetResult(interp, "cannot modify attribute", NULL);
	return TCL_ERROR;

      case TCLDOM_NODE_NODEVALUE:

	evPtr = TclDOMNewEvent(interp, Tcl_NewStringObj("DOMCharacterDataModified", -1));
	newobjv[0] = Tcl_NewStringObj("event", -1);
	newobjv[1] = Tcl_NewStringObj("initMutationEvent", -1);
	newobjv[2] = evPtr;
	newobjv[3] = Tcl_NewStringObj("DOMCharacterDataModified", -1);
	newobjv[4] = Tcl_NewIntObj(1);
	newobjv[5] = Tcl_NewIntObj(0);
	newobjv[6] = Tcl_NewObj();
	newobjv[7] = Tcl_NewStringObj(xmlNodeGetContent(nodePtr), -1);
	newobjv[8] = objv[1];
	newobjv[9] = Tcl_NewObj();
	newobjv[10] = NULL;

	if (TclDOMEventCommand(NULL, interp, 10, newobjv) != TCL_OK) {
	  Tcl_DecrRefCount(evPtr);
	  return TCL_ERROR;
	}

	buf = Tcl_GetStringFromObj(objv[1], &len);
	xmlNodeSetContentLen(nodePtr, buf, len);

	newobjv[0] = Tcl_NewStringObj("node", -1);
	newobjv[1] = Tcl_NewStringObj("dispatchEvent", -1);
	newobjv[2] = nodeObjPtr;
	newobjv[3] = evPtr;
	newobjv[4] = NULL;

	if (TclDOMNodeCommand(NULL, interp, 4, newobjv) != TCL_OK) {
	  Tcl_DecrRefCount(evPtr);
	  return TCL_ERROR;
	}

	Tcl_DecrRefCount(evPtr);

	break;

      case TCLDOM_NODE_CDATASECTION:

	break;
      }

      objc -= 2;
      objv += 2;

    }

    break;

  case TCLDOM_NODE_PARENT:

    if (docPtr) {
      break;
    }

    if (nodePtr->parent) {
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(nodePtr->parent));
    } else {
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromDoc(nodePtr->doc));
    }

    break;

  case TCLDOM_NODE_CHILDREN:

    resultPtr = Tcl_NewListObj(0, NULL);

    if (docPtr) {
      childNodePtr = docPtr->children;
    } else {
      childNodePtr = nodePtr->children;
    }

    while (childNodePtr) {
      Tcl_ListObjAppendElement(interp, resultPtr, TclDOM_CreateObjFromNode(childNodePtr));
      childNodePtr = childNodePtr->next;
    }

    Tcl_SetObjResult(interp, resultPtr);

    break;

  case TCLDOM_NODE_ADDEVENTLISTENER:

    if (objc < 5) {
      Tcl_WrongNumArgs(interp, 2, objv, "addEventListener token type listener ?-usecapture boolean?");
      return TCL_ERROR;
    }

    objc -= 5;
    objv += 5;
    while (objc) {
      if (objc == 1) {
	Tcl_SetResult(interp, "missing value", NULL);
	return TCL_ERROR;
      }
      if (Tcl_GetIndexFromObj(interp, objv[0], NodeCommandAddEventListenerOptions,
			      "option", 0, &option) != TCL_OK) {
	return TCL_ERROR;
      }
      switch ((enum NodeCommandAddEventListenerOptions) option) {
      case TCLDOM_NODE_ADDEVENTLISTENER_USECAPTURE:

	if (Tcl_GetBooleanFromObj(interp, objv[1], &usecapture) != TCL_OK) {
	  return TCL_ERROR;
	}

	break;

      default:
	Tcl_SetResult(interp, "unknown option", NULL);
	return TCL_ERROR;
      }

      objc -= 2;
      objv += 2;
    }

    if (usecapture) {
      tablePtr = &captureListeners;
    } else {
      tablePtr = &bubbleListeners;
    }

    entryPtr = Tcl_CreateHashEntry(tablePtr, (char *) nodePtr, &new);
    if (new) {
      tablePtr = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
      Tcl_InitHashTable(tablePtr, TCL_STRING_KEYS);
      Tcl_SetHashValue(entryPtr, (char *) tablePtr);
    } else {
      tablePtr = (Tcl_HashTable *) Tcl_GetHashValue(entryPtr);
    }

    entryPtr = Tcl_CreateHashEntry(tablePtr, Tcl_GetStringFromObj(typePtr, NULL), &new);
    if (new) {
      Tcl_Obj *listPtr = Tcl_NewListObj(0, NULL);

      Tcl_IncrRefCount(listenerPtr);
      Tcl_IncrRefCount(listPtr);
      Tcl_ListObjAppendElement(interp, listPtr, listenerPtr);
      Tcl_SetHashValue(entryPtr, (char *) listPtr);

    } else {
      Tcl_Obj *listPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
      Tcl_Obj *curPtr;
      int idx, listenerLen, len2;
      char *listenerBuf, *buf2;

      if (Tcl_ListObjLength(interp, listPtr, &len) != TCL_OK) {
	Tcl_SetResult(interp, "internal error - bad list", NULL);
	return TCL_ERROR;
      }
      listenerBuf = Tcl_GetStringFromObj(listenerPtr, &listenerLen);
      for (idx = 0; idx < len; idx++) {
	Tcl_ListObjIndex(interp, listPtr, idx, &curPtr);
	buf2 = Tcl_GetStringFromObj(curPtr, &len2);
	if (listenerLen != len2 || 
	    strncmp(listenerBuf, buf2, listenerLen)) {
	  continue;
	}
      }

      Tcl_ListObjReplace(interp, listPtr, idx, 1, 1, &listenerPtr);
    }

    break;

  case TCLDOM_NODE_REMOVEEVENTLISTENER:

    if (objc < 5) {
      Tcl_WrongNumArgs(interp, 2, objv, "removeEventListener token type listener ?-usecapture boolean?");
      return TCL_ERROR;
    } else {

      objc -= 4;
      while (objc) {
	if (Tcl_GetIndexFromObj(interp, objv[0], NodeCommandAddEventListenerOptions,
				"option", 0, &option) != TCL_OK) {
	  return TCL_ERROR;
	}
	switch ((enum NodeCommandAddEventListenerOptions) option) {
	case TCLDOM_NODE_ADDEVENTLISTENER_USECAPTURE:

	  if (Tcl_GetBooleanFromObj(interp, objv[1], &usecapture) != TCL_OK) {
	    return TCL_ERROR;
	  }

	  break;

	default:
	  Tcl_SetResult(interp, "unknown option", NULL);
	  return TCL_ERROR;
	}
      }
    }

    if (usecapture) {
      tablePtr = &captureListeners;
    } else {
      tablePtr = &bubbleListeners;
    }

    entryPtr = Tcl_FindHashEntry(tablePtr, (char *) nodePtr);
    if (entryPtr) {
      tablePtr = (Tcl_HashTable *) Tcl_GetHashValue(entryPtr);

      entryPtr = Tcl_FindHashEntry(tablePtr, Tcl_GetStringFromObj(typePtr, NULL));
      if (entryPtr) {
	Tcl_Obj *listPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
	Tcl_Obj *curPtr;
	int idx, listenerLen, len2;
	char *listenerBuf, *buf2;

	if (Tcl_ListObjLength(interp, listPtr, &len) != TCL_OK) {
	  Tcl_SetResult(interp, "internal error - bad list", NULL);
	  return TCL_ERROR;
	}
	listenerBuf = Tcl_GetStringFromObj(listenerPtr, &listenerLen);
	for (idx = 0; idx < len; idx++) {
	  Tcl_ListObjIndex(interp, listPtr, idx, &curPtr);
	  buf2 = Tcl_GetStringFromObj(curPtr, &len2);
	  if (listenerLen != len2 || 
	      strncmp(listenerBuf, buf2, listenerLen)) {
	    continue;
	  }
	}

	if (idx == len) {
	  Tcl_SetResult(interp, "no listener registered", NULL);
	  return TCL_ERROR;
	} else {
	  Tcl_ListObjReplace(interp, listPtr, idx, 1, 0, NULL);
	}
      } else {
	Tcl_SetResult(interp, "no listeners registered", NULL);
	return TCL_ERROR;
      }
    } else {
      Tcl_SetResult(interp, "no listeners registered", NULL);
      return TCL_ERROR;
    }

    break;

  case TCLDOM_NODE_DISPATCHEVENT:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "dispatchEvent token event");
      return TCL_ERROR;
    } else if (!nodePtr) {
      /*
       * Events are not dispatched on the root node.
       * This behaviour may not follow the standard,
       * so review later.
       */
      return TCL_OK;
    } else {
      TclDOMEvent *event;
      char *phase;
      Tcl_Obj *pathPtr, *listenerObj;
      int idx, cancelable;

      if (TclDOM_GetEventFromObj(interp, objv[3], &event) != TCL_OK) {
	return TCL_ERROR;
      }
      evPtr = objv[3];

      phase = Tcl_GetStringFromObj(event->eventPhase, &len);

      if (!len) {
	/*
	 * This is the initial dispatch of the event.
	 * First trigger any capturing event listeners
	 * Starting from the root, proceed downward
	 */

	Tcl_SetStringObj(event->eventPhase, "capturing_phase", -1);
	event->target = objv[2];

	pathPtr = TclDOMGetPath(interp, nodePtr);
	Tcl_ListObjLength(interp, pathPtr, &len);
	Tcl_ListObjReplace(interp, pathPtr, len - 1, 1, 0, NULL);
	Tcl_ListObjReplace(interp, pathPtr, 0, 1, 0, NULL);
	Tcl_ListObjLength(interp, pathPtr, &len);
	for (idx = 0; idx < len; idx++) {
	  Tcl_Obj *ancestorObjPtr, *captureListPtr;
	  xmlNodePtr ancestorPtr;
	  int captureIdx, captureLen, cancelable;

	  Tcl_ListObjIndex(interp, pathPtr, idx, &ancestorObjPtr);
	  event->currentNode = ancestorObjPtr;
	  Tcl_IncrRefCount(ancestorObjPtr);
	  if (TclDOM_GetNodeFromObj(interp, ancestorObjPtr, &ancestorPtr) != TCL_OK) {
	    Tcl_SetResult(interp, "cannot find ancestor node \"", NULL);
	    Tcl_AppendResult(interp, Tcl_GetStringFromObj(ancestorObjPtr, NULL), "\"", NULL);
	    return TCL_ERROR;
	  }

	  if (TriggerEventListeners(interp, &captureListeners, ancestorPtr, evPtr) != TCL_OK) {
	    return TCL_ERROR;
	  }

	  /*
	   * A listener may stop propagation,
	   * but we check here to let all of the
	   * listeners at that level complete.
	   */

	  if (Tcl_GetBooleanFromObj(interp, event->cancelable, &cancelable) != TCL_OK) {
	    return TCL_ERROR;
	  }
	  if (cancelable && event->stopPropagation) {
	    Tcl_DecrRefCount(ancestorObjPtr);
	    break;
	  }

	  Tcl_DecrRefCount(ancestorObjPtr);

	}

	/* Prepare for the next phase */

	Tcl_SetStringObj(event->eventPhase, "at_target", -1);
      }

      event->currentNode = nodeObjPtr;

      if (TriggerEventListeners(interp, &bubbleListeners, nodePtr, evPtr) != TCL_OK) {
	return TCL_ERROR;
      }

      Tcl_SetStringObj(event->eventPhase, "bubbling_phase", -1);

      if (Tcl_GetBooleanFromObj(interp, event->cancelable, &cancelable) != TCL_OK) {
	return TCL_ERROR;
      }
      if (cancelable && event->stopPropagation) {
	/* Do no more */
      } else if (nodePtr->parent && nodePtr->parent != (xmlNodePtr) nodePtr->doc) {
	Tcl_Obj *newobjv[5];

	newobjv[0] = Tcl_NewStringObj("node", -1);
	newobjv[1] = Tcl_NewStringObj("dispatchEvent", -1);
	newobjv[2] = TclDOM_CreateObjFromNode(nodePtr->parent);
	newobjv[3] = evPtr;
	newobjv[4] = NULL;
	return TclDOMNodeCommand(dummy, interp, 4, newobjv);
      }

      event->dispatched = 1;

    }

    break;

  default:
    Tcl_SetResult(interp, "not yet implemented", NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMElementCommand --
 *
 *  Implements dom::libxml2::element command.
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMElementCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int method, option;
  xmlNodePtr nodePtr;
  Tcl_Obj *objPtr;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "method ?args...?");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], ElementCommandMethods,
			  "method", 0, &method) != TCL_OK) {
    return TCL_ERROR;
  }

  if (TclDOM_GetNodeFromObj(interp, objv[2], &nodePtr) != TCL_OK) {
    return TCL_ERROR;
  }

  Tcl_SetResult(interp, "element command not yet implemented", NULL);
  return TCL_ERROR;

  switch ((enum ElementCommandMethods) method) {

  case TCLDOM_ELEMENT_CGET:
    break;

  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMEventCommand --
 *
 *  Implements dom::libxml2::event command.
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMEventCommand (clientData, interp, objc, objv)
     ClientData clientData;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int method, option;
  TclDOMEvent *eventPtr;
  xmlNodePtr nodePtr;
  Tcl_Obj *eventObj, *objPtr, *nodeObj;
  Tcl_Obj **argPtr;
  Tcl_Obj *newobjv[10];

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "method ?args...?");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], EventCommandMethods,
			  "method", 0, &method) != TCL_OK) {
    return TCL_ERROR;
  }

  switch ((enum EventCommandMethods) method) {

  case TCLDOM_EVENT_CGET:
    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "cget event option");
      return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[3], EventCommandOptions,
			    "option", 0, &option) != TCL_OK) {
      return TCL_ERROR;
    }

    if (clientData) {
      eventObj = (Tcl_Obj *) clientData;
      eventPtr = (TclDOMEvent *) eventObj->internalRep.otherValuePtr;
    } else {
      eventObj = objv[2];
      if (TclDOM_GetEventFromObj(interp, objv[2], &eventPtr) != TCL_OK) {
	return TCL_ERROR;
      }
    }

    switch ((enum EventCommandOptions) option) {
    case TCLDOM_EVENT_ALTKEY:
      Tcl_SetObjResult(interp, eventPtr->altKey);
      break;
    case TCLDOM_EVENT_ATTRNAME:
      Tcl_SetObjResult(interp, eventPtr->attrName);
      break;
    case TCLDOM_EVENT_BUBBLES:
      Tcl_SetObjResult(interp, eventPtr->bubbles);
      break;
    case TCLDOM_EVENT_BUTTON:
      Tcl_SetObjResult(interp, eventPtr->button);
      break;
    case TCLDOM_EVENT_CANCELABLE:
      Tcl_SetObjResult(interp, eventPtr->cancelable);
      break;
    case TCLDOM_EVENT_CLIENTX:
      Tcl_SetObjResult(interp, eventPtr->clientX);
      break;
    case TCLDOM_EVENT_CLIENTY:
      Tcl_SetObjResult(interp, eventPtr->clientY);
      break;
    case TCLDOM_EVENT_CTRLKEY:
      Tcl_SetObjResult(interp, eventPtr->ctrlKey);
      break;
    case TCLDOM_EVENT_CURRENTNODE:
      Tcl_SetObjResult(interp, eventPtr->currentNode);
      break;
    case TCLDOM_EVENT_DETAIL:
      Tcl_SetObjResult(interp, eventPtr->detail);
      break;
    case TCLDOM_EVENT_EVENTPHASE:
      Tcl_SetObjResult(interp, eventPtr->eventPhase);
      break;
    case TCLDOM_EVENT_METAKEY:
      Tcl_SetObjResult(interp, eventPtr->metaKey);
      break;
    case TCLDOM_EVENT_NEWVALUE:
      Tcl_SetObjResult(interp, eventPtr->newValue);
      break;
    case TCLDOM_EVENT_PREVVALUE:
      Tcl_SetObjResult(interp, eventPtr->prevValue);
      break;
    case TCLDOM_EVENT_RELATEDNODE:
      Tcl_SetObjResult(interp, eventPtr->relatedNode);
      break;
    case TCLDOM_EVENT_SCREENX:
      Tcl_SetObjResult(interp, eventPtr->screenX);
      break;
    case TCLDOM_EVENT_SCREENY:
      Tcl_SetObjResult(interp, eventPtr->screenY);
      break;
    case TCLDOM_EVENT_SHIFTKEY:
      Tcl_SetObjResult(interp, eventPtr->shiftKey);
      break;
    case TCLDOM_EVENT_TARGET:
      Tcl_SetObjResult(interp, eventPtr->target);
      break;
    case TCLDOM_EVENT_TIMESTAMP:
      Tcl_SetObjResult(interp, eventPtr->timeStamp);
      break;
    case TCLDOM_EVENT_TYPE:
      Tcl_SetObjResult(interp, eventPtr->type);
      break;
    case TCLDOM_EVENT_VIEW:
      Tcl_SetObjResult(interp, eventPtr->view);
      break;
    default:
      Tcl_SetResult(interp, "unknown option", NULL);
      return TCL_ERROR;
    }

    break;

  case TCLDOM_EVENT_CONFIGURE:
    if (objc < 2) {
      Tcl_WrongNumArgs(interp, 3, objv, "configure option ?value?");
      return TCL_ERROR;
    }

    /* No event options are writable */
    Tcl_SetResult(interp, "option cannot be modified", NULL);
    return TCL_ERROR;

    break;

  case TCLDOM_EVENT_STOPPROPAGATION:

    if (clientData) {
      eventObj = (Tcl_Obj *) clientData;
      eventPtr = (TclDOMEvent *) eventObj->internalRep.otherValuePtr;
    } else {
      eventObj = objv[2];
      if (TclDOM_GetEventFromObj(interp, objv[2], &eventPtr) != TCL_OK) {
	return TCL_ERROR;
      }
    }

    if (objc != 0) {
      Tcl_WrongNumArgs(interp, 1, objv, "");
      return TCL_ERROR;
    }

    eventPtr->stopPropagation = 1;

    break;

  case TCLDOM_EVENT_PREVENTDEFAULT:

    if (clientData) {
      eventObj = (Tcl_Obj *) clientData;
      eventPtr = (TclDOMEvent *) eventObj->internalRep.otherValuePtr;
    } else {
      eventObj = objv[2];
      if (TclDOM_GetEventFromObj(interp, objv[2], &eventPtr) != TCL_OK) {
	return TCL_ERROR;
      }
    }

    if (objc != 0) {
      Tcl_WrongNumArgs(interp, 2, objv, "");
      return TCL_ERROR;
    }

    eventPtr->preventDefault = 1;

    break;

  case TCLDOM_EVENT_INITEVENT:

    if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "type bubbles cancelable");
      return TCL_ERROR;
    }

    if (clientData) {
      eventObj = (Tcl_Obj *) clientData;
      eventPtr = (TclDOMEvent *) eventObj->internalRep.otherValuePtr;
      argPtr = (Tcl_Obj **) &objv[2];
      objc -= 2;
    } else {
      eventObj = objv[2];
      if (TclDOM_GetEventFromObj(interp, objv[2], &eventPtr) != TCL_OK) {
	return TCL_ERROR;
      }
      argPtr = (Tcl_Obj **) &objv[3];
      objc -= 3;
    }

    if (eventPtr->dispatched) {
      Tcl_SetResult(interp, "event has been dispatched", NULL);
      return TCL_ERROR;
    }

    Tcl_DecrRefCount(eventPtr->type);
    eventPtr->type = argPtr[0];
    Tcl_IncrRefCount(eventPtr->type);
    Tcl_DecrRefCount(eventPtr->bubbles);
    eventPtr->bubbles = argPtr[1];
    Tcl_IncrRefCount(eventPtr->bubbles);
    Tcl_DecrRefCount(eventPtr->cancelable);
    eventPtr->cancelable = argPtr[2];
    Tcl_IncrRefCount(eventPtr->cancelable);

    break;

  case TCLDOM_EVENT_INITUIEVENT:

    if (objc < 4 || objc > 5) {
      Tcl_WrongNumArgs(interp, 2, objv, "type bubbles cancelable view ?detail?");
      return TCL_ERROR;
    }

    if (clientData) {
      eventObj = (Tcl_Obj *) clientData;
      eventPtr = (TclDOMEvent *) eventObj->internalRep.otherValuePtr;
      argPtr = (Tcl_Obj **) &objv[2];
      objc -= 2;
    } else {
      eventObj = objv[2];
      if (TclDOM_GetEventFromObj(interp, objv[2], &eventPtr) != TCL_OK) {
	return TCL_ERROR;
      }
      argPtr = (Tcl_Obj **) &objv[3];
      objc -= 3;
    }

    if (eventPtr->dispatched) {
      Tcl_SetResult(interp, "event has been dispatched", NULL);
      return TCL_ERROR;
    }

    Tcl_DecrRefCount(eventPtr->type);
    eventPtr->type = argPtr[0];
    Tcl_IncrRefCount(eventPtr->type);
    Tcl_DecrRefCount(eventPtr->bubbles);
    eventPtr->bubbles = argPtr[1];
    Tcl_IncrRefCount(eventPtr->bubbles);
    Tcl_DecrRefCount(eventPtr->cancelable);
    eventPtr->cancelable = argPtr[2];
    Tcl_IncrRefCount(eventPtr->cancelable);

    Tcl_DecrRefCount(eventPtr->view);
    eventPtr->view = argPtr[3];
    Tcl_IncrRefCount(eventPtr->view);
    if (objc == 5) {
      Tcl_DecrRefCount(eventPtr->detail);
      eventPtr->detail = argPtr[4];
      Tcl_IncrRefCount(eventPtr->detail);
    }

    break;

  case TCLDOM_EVENT_INITMOUSEEVENT:

    if (objc != 18) {
      Tcl_WrongNumArgs(interp, 2, objv, "type bubbles cancelable view detail screenX screenY clientX clientY ctrlKey altKey shiftKey metaKey button relatedNode");
      return TCL_ERROR;
    }

    if (clientData) {
      eventObj = (Tcl_Obj *) clientData;
      eventPtr = (TclDOMEvent *) eventObj->internalRep.otherValuePtr;
      argPtr = (Tcl_Obj **) &objv[2];
      objc -= 2;
    } else {
      eventObj = objv[2];
      if (TclDOM_GetEventFromObj(interp, objv[2], &eventPtr) != TCL_OK) {
	return TCL_ERROR;
      }
      argPtr = (Tcl_Obj **) &objv[3];
      objc -= 3;
    }

    if (eventPtr->dispatched) {
      Tcl_SetResult(interp, "event has been dispatched", NULL);
      return TCL_ERROR;
    }

    Tcl_DecrRefCount(eventPtr->type);
    eventPtr->type = argPtr[0];
    Tcl_IncrRefCount(eventPtr->type);
    Tcl_DecrRefCount(eventPtr->bubbles);
    eventPtr->bubbles = argPtr[1];
    Tcl_IncrRefCount(eventPtr->bubbles);
    Tcl_DecrRefCount(eventPtr->cancelable);
    eventPtr->cancelable = argPtr[2];
    Tcl_IncrRefCount(eventPtr->cancelable);

    Tcl_DecrRefCount(eventPtr->view);
    eventPtr->view = argPtr[3];
    Tcl_IncrRefCount(eventPtr->view);
    Tcl_DecrRefCount(eventPtr->detail);
    eventPtr->detail = argPtr[4];
    Tcl_IncrRefCount(eventPtr->detail);

    Tcl_DecrRefCount(eventPtr->screenX);
    eventPtr->screenX = argPtr[5];
    Tcl_IncrRefCount(eventPtr->screenX);
    Tcl_DecrRefCount(eventPtr->screenY);
    eventPtr->screenY = argPtr[6];
    Tcl_IncrRefCount(eventPtr->screenY);

    Tcl_DecrRefCount(eventPtr->clientX);
    eventPtr->clientX = argPtr[7];
    Tcl_IncrRefCount(eventPtr->clientX);
    Tcl_DecrRefCount(eventPtr->clientY);
    eventPtr->clientY = argPtr[8];
    Tcl_IncrRefCount(eventPtr->clientY);

    Tcl_DecrRefCount(eventPtr->ctrlKey);
    eventPtr->ctrlKey = argPtr[9];
    Tcl_IncrRefCount(eventPtr->ctrlKey);
    Tcl_DecrRefCount(eventPtr->altKey);
    eventPtr->altKey = argPtr[10];
    Tcl_IncrRefCount(eventPtr->altKey);
    Tcl_DecrRefCount(eventPtr->shiftKey);
    eventPtr->shiftKey = argPtr[11];
    Tcl_IncrRefCount(eventPtr->shiftKey);
    Tcl_DecrRefCount(eventPtr->metaKey);
    eventPtr->metaKey = argPtr[12];
    Tcl_IncrRefCount(eventPtr->metaKey);

    Tcl_DecrRefCount(eventPtr->relatedNode);
    eventPtr->relatedNode = argPtr[13];
    Tcl_IncrRefCount(eventPtr->relatedNode);

    break;

  case TCLDOM_EVENT_INITMUTATIONEVENT:

    if (objc != 10) {
      Tcl_WrongNumArgs(interp, 2, objv, "type bubbles cancelable relatedNode prevValue newValue attrName");
      return TCL_ERROR;
    }

    if (clientData) {
      eventObj = (Tcl_Obj *) clientData;
      eventPtr = (TclDOMEvent *) eventObj->internalRep.otherValuePtr;
      argPtr = (Tcl_Obj **) &objv[2];
      objc -= 2;
    } else {
      eventObj = objv[2];
      if (TclDOM_GetEventFromObj(interp, objv[2], &eventPtr) != TCL_OK) {
	return TCL_ERROR;
      }
      argPtr = (Tcl_Obj **) &objv[3];
      objc -= 3;
    }

    if (eventPtr->dispatched) {
      Tcl_SetResult(interp, "event has been dispatched", NULL);
      return TCL_ERROR;
    }

    Tcl_DecrRefCount(eventPtr->type);
    eventPtr->type = argPtr[0];
    Tcl_IncrRefCount(eventPtr->type);
    Tcl_DecrRefCount(eventPtr->bubbles);
    eventPtr->bubbles = argPtr[1];
    Tcl_IncrRefCount(eventPtr->bubbles);
    Tcl_DecrRefCount(eventPtr->cancelable);
    eventPtr->cancelable = argPtr[2];
    Tcl_IncrRefCount(eventPtr->cancelable);

    Tcl_DecrRefCount(eventPtr->relatedNode);
    eventPtr->relatedNode = argPtr[3];
    Tcl_IncrRefCount(eventPtr->relatedNode);

    Tcl_DecrRefCount(eventPtr->prevValue);
    eventPtr->prevValue = argPtr[4];
    Tcl_IncrRefCount(eventPtr->prevValue);
    Tcl_DecrRefCount(eventPtr->newValue);
    eventPtr->newValue = argPtr[5];
    Tcl_IncrRefCount(eventPtr->newValue);
    Tcl_DecrRefCount(eventPtr->attrName);
    eventPtr->attrName = argPtr[6];
    Tcl_IncrRefCount(eventPtr->attrName);

    break;

  case TCLDOM_EVENT_POSTUIEVENT:

    Tcl_SetResult(interp, "not yet implemented", NULL);
    return TCL_ERROR;

  case TCLDOM_EVENT_POSTMOUSEEVENT:

    Tcl_SetResult(interp, "not yet implemented", NULL);
    return TCL_ERROR;

  case TCLDOM_EVENT_POSTMUTATIONEVENT:

    if (objc < 4) {
      Tcl_WrongNumArgs(interp, 1, objv, "postMutationEvent node type ?args ...?");
      return TCL_ERROR;
    }

    if (TclDOM_GetNodeFromObj(interp, objv[2], &nodePtr) != TCL_OK) {
      return TCL_ERROR;
    }
    nodeObj = objv[2];

    eventObj = TclDOMNewEvent(interp, objv[3]);
    eventPtr = (TclDOMEvent *) eventObj->internalRep.otherValuePtr;

    Tcl_DecrRefCount(eventPtr->bubbles);
    objPtr = Tcl_GetVar2Ex(interp, "::dom::bubbles", Tcl_GetStringFromObj(objv[3], NULL), TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
    if (!objPtr) {
      return TCL_ERROR;
    }
    eventPtr->bubbles = Tcl_DuplicateObj(objPtr);
    Tcl_DecrRefCount(eventPtr->cancelable);
    objPtr = Tcl_GetVar2Ex(interp, "::dom::cancelable", Tcl_GetStringFromObj(objv[3], NULL), TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
    if (!objPtr) {
      return TCL_ERROR;
    }
    eventPtr->cancelable = Tcl_DuplicateObj(objPtr);

    objc -= 4;
    objv += 4;
    while (objc) {

      if (objc == 1) {
	Tcl_SetResult(interp, "value missing", NULL);
	return TCL_ERROR;
      }

      if (Tcl_GetIndexFromObj(interp, objv[0], EventCommandOptions,
			      "option", 0, &option) != TCL_OK) {
	return TCL_ERROR;
      }
      switch ((enum EventCommandOptions) option) {
      case TCLDOM_EVENT_BUBBLES:
	Tcl_DecrRefCount(eventPtr->bubbles);
	eventPtr->bubbles = objv[1];
	Tcl_IncrRefCount(eventPtr->bubbles);
	break;
      case TCLDOM_EVENT_CANCELABLE:
	Tcl_DecrRefCount(eventPtr->cancelable);
	eventPtr->cancelable = objv[1];
	Tcl_IncrRefCount(eventPtr->cancelable);
	break;
      case TCLDOM_EVENT_RELATEDNODE:
	Tcl_DecrRefCount(eventPtr->relatedNode);
	eventPtr->relatedNode = objv[1];
	Tcl_IncrRefCount(eventPtr->relatedNode);
	break;
      case TCLDOM_EVENT_PREVVALUE:
	Tcl_DecrRefCount(eventPtr->prevValue);
	eventPtr->prevValue = objv[1];
	Tcl_IncrRefCount(eventPtr->prevValue);
	break;
      case TCLDOM_EVENT_NEWVALUE:
	Tcl_DecrRefCount(eventPtr->newValue);
	eventPtr->newValue = objv[1];
	Tcl_IncrRefCount(eventPtr->newValue);
	break;
      case TCLDOM_EVENT_ATTRNAME:
	Tcl_DecrRefCount(eventPtr->attrName);
	eventPtr->attrName = objv[1];
	Tcl_IncrRefCount(eventPtr->attrName);
	break;
      default:
	Tcl_SetResult(interp, "bad option", NULL);
	return TCL_ERROR;
      }

      objc -= 2;
      objv += 2;
    }

    newobjv[0] = Tcl_NewStringObj("node", -1);
    newobjv[1] = Tcl_NewStringObj("dispatchEvent", -1);
    newobjv[2] = nodeObj;
    newobjv[3] = eventObj;
    newobjv[4] = NULL;
    if (TclDOMNodeCommand(NULL, interp, 4, newobjv) != TCL_OK) {
      return TCL_ERROR;
    }

    /* TclDOMFreeEvent(eventPtr); */

    break;

  default:

    Tcl_SetResult(interp, "unknown method", NULL);
    return TCL_ERROR;

  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMNewEvent --
 *
 *  Create a Tcl_Obj for an event.
 *
 * Results:
 *  Returns Tcl_Obj*.
 *
 * Side effects:
 *  Allocates object.
 *
 *----------------------------------------------------------------------------
 */

Tcl_Obj *
TclDOMNewEvent (interp, type)
     Tcl_Interp *interp;
     Tcl_Obj *type;
{
  Tcl_Obj *newPtr;
  TclDOMEvent *eventPtr;
  Tcl_Time time;
  Tcl_HashEntry *entry;
  int inew;

  newPtr = Tcl_NewObj();

  newPtr->bytes = Tcl_Alloc(20);
  sprintf(newPtr->bytes, "event%d", eventCntr++);
  newPtr->length = strlen(newPtr->bytes);

  eventPtr = (TclDOMEvent *) Tcl_Alloc(sizeof(TclDOMEvent));
  eventPtr->interp;

  eventPtr->stopPropagation = 0;
  eventPtr->preventDefault = 0;
  eventPtr->dispatched = 0;

  eventPtr->altKey = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->altKey);
  eventPtr->attrName = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->attrName);
  eventPtr->bubbles = Tcl_NewIntObj(1);
  Tcl_IncrRefCount(eventPtr->bubbles);
  eventPtr->button = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->button);
  eventPtr->cancelable = Tcl_NewIntObj(1);
  Tcl_IncrRefCount(eventPtr->cancelable);
  eventPtr->clientX = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->clientX);
  eventPtr->clientY = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->clientY);
  eventPtr->ctrlKey = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->ctrlKey);
  eventPtr->currentNode = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->currentNode);
  eventPtr->detail = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->detail);
  eventPtr->eventPhase = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->eventPhase);
  eventPtr->metaKey = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->metaKey);
  eventPtr->newValue = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->newValue);
  eventPtr->prevValue = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->prevValue);
  eventPtr->relatedNode = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->relatedNode);
  eventPtr->screenX = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->screenX);
  eventPtr->screenY = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->screenY);
  eventPtr->shiftKey = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->shiftKey);
  eventPtr->target = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->target);

  Tcl_GetTime(&time);
  eventPtr->timeStamp = Tcl_NewLongObj(time.sec*1000 + time.usec/1000);
  Tcl_IncrRefCount(eventPtr->timeStamp);

  eventPtr->type = type;
  Tcl_IncrRefCount(eventPtr->type);
  eventPtr->view = Tcl_NewObj();
  Tcl_IncrRefCount(eventPtr->view);

  newPtr->internalRep.otherValuePtr = (VOID *) eventPtr;
  newPtr->typePtr = &TclDOM_EventObjType;

  entry = Tcl_CreateHashEntry(&events, newPtr->bytes, &inew);
  Tcl_SetHashValue(entry, (void *) eventPtr);
  Tcl_IncrRefCount(newPtr);

  Tcl_CreateObjCommand(interp, newPtr->bytes, TclDOMEventCommand, (ClientData) eventPtr, NULL);

  return newPtr;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOM_CreateObjFromDoc --
 *
 *  Create a Tcl_Obj to wrap a document.
 *
 * Results:
 *  Returns Tcl_Obj*.
 *
 * Side effects:
 *  Allocates object.
 *
 *----------------------------------------------------------------------------
 */

Tcl_Obj *
TclDOM_CreateObjFromDoc (docPtr)
  xmlDocPtr docPtr;
{

  /*
   * This document may have already been wrapped by a Tcl_Object.
   */

  if (docPtr->_private) {
    Tcl_IncrRefCount((Tcl_Obj *) docPtr->_private);
    return (Tcl_Obj *) docPtr->_private;
  } else {
    Tcl_Obj *objPtr;
    Tcl_HashEntry *entry;
    int new;

    objPtr = Tcl_NewObj();
    Tcl_IncrRefCount(objPtr);
    objPtr->internalRep.otherValuePtr = (VOID *) docPtr;
    objPtr->typePtr = &TclDOM_DocObjType;
    objPtr->bytes = Tcl_Alloc(20);
    sprintf(objPtr->bytes, "doc%d", docCntr++);
    objPtr->length = strlen(objPtr->bytes);

    entry = Tcl_CreateHashEntry(&documents, objPtr->bytes, &new);
    Tcl_SetHashValue(entry, docPtr);

    docPtr->_private = (void *) objPtr;

    return objPtr;
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOM_CreateObjFromNode --
 *
 *  Create a Tcl_Obj to wrap a tree node.
 *
 *  NB. We could get alot fancier in generating a
 *  string rep, eg. to allow mapping from a string
 *  rep back to an xmlNodePtr.  However, this will
 *  use alot more memory since there may be many
 *  nodes.  So, the application will have to
 *  keep hold of the internal rep.
 *
 * Results:
 *  Returns Tcl_Obj*.
 *
 * Side effects:
 *  Allocates object.
 *
 *----------------------------------------------------------------------------
 */

Tcl_Obj *
TclDOM_CreateObjFromNode (nodePtr)
  xmlNodePtr nodePtr;
{
  if (nodePtr->_private) {
    Tcl_IncrRefCount((Tcl_Obj *) nodePtr->_private);
    return (Tcl_Obj *) nodePtr->_private;
  } else {
    Tcl_Obj *objPtr;
    Tcl_HashEntry *entry;
    int new;

    objPtr = Tcl_NewObj();
    Tcl_IncrRefCount(objPtr);
    objPtr->internalRep.otherValuePtr = (VOID *) nodePtr;
    objPtr->typePtr = &TclDOM_NodeObjType;
    objPtr->bytes = Tcl_Alloc(20);
    sprintf(objPtr->bytes, "domnode%d", nodeCntr++);
    objPtr->length = strlen(objPtr->bytes);

    nodePtr->_private = (void *) objPtr;

    entry = Tcl_CreateHashEntry(&nodes, objPtr->bytes, &new);
    if (!new) {
      /* internal error: this should never occur */
    }
    Tcl_SetHashValue(entry, (void *) nodePtr);

    return objPtr;
  }
}

/*
 *	Manage TclDOM objects
 */

/*
 *----------------------------------------------------------------------------
 *
 * TclDOM_GetDocFromObj --
 *
 *  Gets an xmlDocPtr from a Tcl_Obj.
 *
 * Results:
 *  Returns success code.
 *
 * Side effects:
 *  None.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOM_GetDocFromObj (interp, objPtr, docPtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
     xmlDocPtr *docPtr;
{

  if (objPtr->typePtr == &TclDOM_DocObjType) {
    *docPtr = (xmlDocPtr) objPtr->internalRep.otherValuePtr;
  } else if (TclDOM_DocSetFromAny(interp, objPtr) == TCL_OK) {
    *docPtr = (xmlDocPtr) objPtr->internalRep.otherValuePtr;
  } else {
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOM_GetNodeFromObj --
 *
 *  Gets an xmlNodePtr from a Tcl_Obj.
 *
 * Results:
 *  Returns success code.
 *
 * Side effects:
 *  None.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOM_GetNodeFromObj (interp, objPtr, nodePtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
     xmlNodePtr *nodePtr;
{

  if (objPtr->typePtr == &TclDOM_NodeObjType) {
     *nodePtr = objPtr->internalRep.otherValuePtr;
  } else if (TclDOM_NodeSetFromAny(interp, objPtr) == TCL_OK) {
    *nodePtr = objPtr->internalRep.otherValuePtr;
  } else {
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMNewDoc --
 *
 *  Creates a new xmlDocPtr and wraps it in a Tcl_Obj.
 *
 * Results:
 *  Returns a *TclObj
 *
 * Side effects:
 *  Objects allocated.
 *
 *----------------------------------------------------------------------------
 */

Tcl_Obj *
TclDOMNewDoc(interp)
     Tcl_Interp *interp;
{
  Tcl_Obj *objPtr;
  xmlDocPtr new;
  Tcl_HashEntry *entry;
  int inew;

  new = xmlNewDoc("1.0");
  if (!new) {
    return NULL;
  }

  objPtr = Tcl_NewObj();

  objPtr->internalRep.otherValuePtr = (VOID *) new;
  objPtr->typePtr = &TclDOM_DocObjType;
  objPtr->bytes = Tcl_Alloc(20);
  sprintf(objPtr->bytes, "doc%d", docCntr++);
  objPtr->length = strlen(objPtr->bytes);

  new->_private = (void *) objPtr;

  entry = Tcl_CreateHashEntry(&documents, objPtr->bytes, &inew);
  Tcl_SetHashValue(entry, (void *) new);

  Tcl_IncrRefCount(objPtr);

  return objPtr;
}

int
TclDOM_DocSetFromAny(interp, objPtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
{
  Tcl_HashEntry *entryPtr;

  entryPtr = Tcl_FindHashEntry(&documents, Tcl_GetStringFromObj(objPtr, NULL));

  if (entryPtr) {

    if (objPtr->typePtr != NULL && objPtr->typePtr->freeIntRepProc != NULL) {
      objPtr->typePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.otherValuePtr = Tcl_GetHashValue(entryPtr);
    objPtr->typePtr = &TclDOM_DocObjType;

  } else {
    Tcl_SetResult(interp, "no such DOM document - try parsing first", NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

void
TclDOM_DocUpdate(objPtr)
     Tcl_Obj *objPtr;
{
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;

  entryPtr = Tcl_FirstHashEntry(&documents, &search);
  while (entryPtr && objPtr->internalRep.otherValuePtr != Tcl_GetHashValue(entryPtr)) {
    entryPtr = Tcl_NextHashEntry(&search);
  }
  Tcl_InvalidateStringRep(objPtr);
  if (entryPtr == NULL) {
    objPtr->bytes = NULL;
    objPtr->length = 0;
  } else {
    objPtr->length = strlen(Tcl_GetHashKey(&documents, entryPtr));
    objPtr->bytes = Tcl_Alloc(objPtr->length + 1);
    strcpy(objPtr->bytes, Tcl_GetHashKey(&documents, entryPtr));
  }
}

void
TclDOM_DocDup(srcPtr, dstPtr)
     Tcl_Obj *srcPtr;
     Tcl_Obj *dstPtr;
{
  xmlDocPtr srcDoc, dstDoc;
  Tcl_HashEntry *entry;
  int new;

  srcDoc = (xmlDocPtr) srcPtr->internalRep.otherValuePtr;
  dstDoc = xmlCopyDoc(srcDoc, 1);
  if (dstDoc == NULL) {
    return;
  }

  /*
   * xmlCopyDoc doesn't copy some of the fields.
   */

  if (srcDoc->URL) {
    dstDoc->URL = Tcl_Alloc(strlen(srcDoc->URL) + 1);
    strcpy((char *) dstDoc->URL, srcDoc->URL);
  }

  Tcl_InvalidateStringRep(dstPtr);

  dstDoc->_private = (void *) dstPtr;
  dstPtr->internalRep.otherValuePtr = (VOID *) dstDoc;

  dstPtr->bytes = Tcl_Alloc(20);
  sprintf(dstPtr->bytes, "doc%d", docCntr++);
  dstPtr->length = strlen(dstPtr->bytes);

  entry = Tcl_CreateHashEntry(&documents, dstPtr->bytes, &new);
  Tcl_SetHashValue(entry, (void *) dstDoc);

}

void
TclDOM_DocFree(objPtr)
     Tcl_Obj *objPtr;
{
  /*
  Tcl_DecrRefCount(objPtr);
  */
}

int
TclDOM_NodeSetFromAny(interp, objPtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
{
  Tcl_HashEntry *entry;

  entry = Tcl_FindHashEntry(&nodes, Tcl_GetStringFromObj(objPtr, NULL));
  if (entry) {

    if (objPtr->typePtr != NULL && objPtr->typePtr->freeIntRepProc != NULL) {
      objPtr->typePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.otherValuePtr = Tcl_GetHashValue(entry);
    objPtr->typePtr = &TclDOM_NodeObjType;

    return TCL_OK;

  } else {
    Tcl_SetResult(interp, "not a DOM node", NULL);
    return TCL_ERROR;
  }
}

void
TclDOM_NodeUpdate(objPtr)
     Tcl_Obj *objPtr;
{
  /* Search through the nodes hash table for this
   * node and use its key as the string rep.
   */
}

/* This operation is not supported */
void
TclDOM_NodeDup(srcPtr, dstPtr)
     Tcl_Obj *srcPtr;
     Tcl_Obj *dstPtr;
{
  Tcl_InvalidateStringRep(dstPtr);
}

/* This operation does nothing - app should use dom::node interface */
void
TclDOM_NodeFree(objPtr)
     Tcl_Obj *objPtr;
{
}

int
TclDOM_EventSetFromAny(interp, objPtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
{
  Tcl_HashEntry *entry;

  entry = Tcl_FindHashEntry(&events, Tcl_GetStringFromObj(objPtr, NULL));
  if (entry) {

    if (objPtr->typePtr != NULL && objPtr->typePtr->freeIntRepProc != NULL) {
      objPtr->typePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.otherValuePtr = Tcl_GetHashValue(entry);
    objPtr->typePtr = &TclDOM_EventObjType;
  } else {
    Tcl_SetResult(interp, "unable to find event", NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

void
TclDOM_EventUpdate(objPtr)
     Tcl_Obj *objPtr;
{
  /* TODO: Search for this event in the events hash table. */
}

/* This operation is not supported */
void
TclDOM_EventDup(srcPtr, dstPtr)
     Tcl_Obj *srcPtr;
     Tcl_Obj *dstPtr;
{
  Tcl_InvalidateStringRep(dstPtr);
}

/* This operation does nothing - app should use dom::event interface */
void
TclDOM_EventFree(objPtr)
     Tcl_Obj *objPtr;
{
  /*
   * Tcl objects are simply pointers to event structures in
   * the events hash table, so free'ing an event-typed object
   * is of no consequence.  Events must be destroyed using the
   * dom::destroy command.
   *
   * An alternative would be to maintain a reference count.
   */
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMGetPath --
 *
 *  Constructs a list of ancestor nodes.
 *
 * Results:
 *  Returns list as a Tcl_Obj.
 *
 * Side effects:
 *  Allocates Tcl_Obj structures.
 *
 *----------------------------------------------------------------------------
 */

static Tcl_Obj *
TclDOMGetPath (interp, nodePtr)
     Tcl_Interp *interp;
     xmlNodePtr nodePtr;
{
  Tcl_Obj *listPtr, *resultPtr;
  Tcl_Obj *objv[2];

  if (nodePtr) {
    objv[0] = TclDOM_CreateObjFromNode(nodePtr);
    objv[1] = NULL;

    listPtr = Tcl_NewListObj(1, objv);
    if (nodePtr->parent) {
      resultPtr = TclDOMGetPath(interp, nodePtr->parent);
      Tcl_ListObjAppendList(interp, resultPtr, listPtr);
    } else {
      resultPtr = listPtr;
    }
    return resultPtr;
  } else {
    return Tcl_NewObj();
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOM_GetEventFromObj --
 *
 *  Returns TclDOMEvent pointer from Tcl object.
 *
 * Results:
 *  Returns success code and assigns event pointer.
 *
 * Side effects:
 *  May invalidate object's internal rep.
 *
 *----------------------------------------------------------------------------
 */

static int
TclDOM_GetEventFromObj (interp, objPtr, eventPtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
     TclDOMEvent **eventPtr;
{
  if (!objPtr) {
    Tcl_SetResult(interp, "null object", NULL);
    return TCL_ERROR;
  }
  if (objPtr->typePtr == &TclDOM_EventObjType) {
    *eventPtr = (TclDOMEvent *) objPtr->internalRep.otherValuePtr;
  } else if (TclDOM_EventSetFromAny(interp, objPtr) == TCL_OK) {
    Tcl_SetResult(interp, "about to set eventPtr", NULL);
    return TCL_ERROR;
    *eventPtr = (TclDOMEvent *) objPtr->internalRep.otherValuePtr;
  } else {
    return TCL_ERROR;
  }
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMGenericError --
 *
 *  Handler for parse errors.
 *
 * Results:
 *  Stores error message.
 *
 * Side effects:
 *  Transform will return error condition.
 *
 *----------------------------------------------------------------------------
 */

static void
TclDOMGenericError (void *ctx, const char *msg, ...)
{
  va_list args;
  char buf[2048];
  int len;

  errorInfo.code = TCL_ERROR;

  if (!errorInfo.msg) {
    errorInfo.msg = Tcl_NewObj();
    Tcl_IncrRefCount(errorInfo.msg);
  }

  va_start(args,msg);
  len = vsnprintf(buf, 2047, msg, args);
  va_end(args);

  Tcl_AppendToObj(errorInfo.msg, buf, len);

}

